{$F+
.po2}
UNIT fitter;

interface

const
        mm          = 8;
		mnvpp       = 3;
        mnp         = 512;
type
        vector      = array[1..mm+1]  of extended;
        datarow     = array[1..mnvpp] of extended;
        dataset     = array[1..mnp]   of datarow;
        functype    = function(x : vector; d : datarow) : extended;
var
        exitsave    : pointer;
        dout        : text;

procedure simplex(
               f           : functype;
               trace       : boolean;
               nvpp        : byte;
               m           : byte;
               np          : integer;
               maxiter     : longint;
               maxerr      : extended;
               data        : dataset;
               var mean    : vector;
               var niter   : longint);

procedure jknife(
               f           : functype;   {function}
               trace       : boolean;
               nvpp        : byte;
               m           : byte;
               np          : integer;
               maxiter     : longint;
               maxerr      : extended;
               data        : dataset;
               var mean    : vector;     {est.value}
               var error   : vector;     {est.error}
               var niter   : longint);   {number successes}

procedure bstrap(
               f           : functype;   {function}
               trace       : boolean;
               nvpp        : byte;
               m           : byte;
               np          : integer;
               maxiter     : longint;
               maxerr      : extended;
               data        : dataset;
               var mean    : vector;     {est.value}
               var error   : vector;     {est.error}
               var niter   : longint);   {number tests/successes}

procedure mcarlo(
               f           : functype;   {function}
               trace       : boolean;
               nvpp        : byte;
               m           : byte;
               np          : integer;
               maxiter     : longint;
               maxerr      : extended;
               data        : dataset;
               var mean    : vector;     {est.value}
               var error   : vector;     {est.error}
               var niter   : longint);   {number tests/successes}

implementation

procedure simplex;

const   track = false;
        alfa  = 1.0;
        beta  = 0.5;
        gamma = 2.0;
        root2 = 1.414214;
        stepfrac = 0.05;
var
        n,i,j,h,l,o             : byte;
        error                   : extended;
        next,center,p,q,step    : vector;
        simp                    : array[1..mm+1] of vector;

procedure myexit;
  begin
  niter := maxiter;
  writeln('abnormal exit from simplex procedure, error ',exitcode);
  erroraddr := nil;
  exitproc  := exitsave;
  end;

procedure sum_of_residuals (var x:vector);
  var i : byte;
  begin
  x[n] := 0.0;
  for i := 1 to np do x[n] := x[n] + sqr(f(x,data[i]) - data[i,nvpp])
  end;

procedure new_vertex;
  var i : byte;
  begin
  if track then
    begin
    for i := 1 to m+1 do
      write(dout,next[i]:13,' ');
    writeln(dout);
    end;
  if trace then write(niter:4);
  for i := 1 to n do begin
    simp[h,i] := next[i];
    if trace then write(' ',next[i]:14)
    end;
  if trace then writeln
  end;

procedure order;
  var     i  : byte;
  begin
  for i := 1 to n do begin
    if simp[i,n] < simp[l,n] then l := i;
    if simp[i,n] > simp[h,n] then h := i
    end;
  o := l;
  for i := 1 to n do if i <> h then if simp[i,n] > simp[o,n] then o := i
  end;

begin                                                          {simplex}
if track then
  begin
  assign(dout,'simp.res');
{$I-}
  erase(dout);
  rewrite(dout);
  close(dout);
  i := ioresult;
{$I+}
  append(dout);
  end;
exitsave := exitproc;
exitproc := @myexit;
n := m + 1;
for i := 1 to m do simp[1,i] := mean[i];
for i := 1 to m do step[i] := mean[i]*stepfrac;
for i := 1 to m do begin
  p[i] := step[i] * (sqrt(n) + m - 1) / (m * root2);
  q[i] := step[i] * (sqrt(n) - 1) / (m * root2)
  end;
sum_of_residuals(simp[1]);
for i := 2 to n do begin
  for j := 1 to m do simp[i,j] := simp[1,j] + q[j];
  simp[i,i-1] := simp[1,i-1] + p[i-1];
  sum_of_residuals(simp[i])
  end;
l := 1; h := 1; o := 1;
order;
if trace then begin
  writeln(' starting simplex');
  for j := 1 to n do begin
    write(' simp[',j:1,']');
    for i := 1 to n do write(' ',simp[j,i]:14);
    writeln;
    end;
  end;
niter :=0;
repeat
  niter := succ(niter);
  for i := 1 to n do center[i] := 0.0;
  for i := 1 to n do if i <> h then
    for j := 1 to m do center[j] := center[j] + simp[i,j];
  for i := 1 to n do begin
    center[i] := center[i] / m;
    next[i] := (1.0 + alfa) * center[i] - alfa * simp[h,i]
    end;
  sum_of_residuals(next);
  if next[n] <= simp[l,n] then begin
    if trace then write(' r');
    new_vertex;
    for i := 1 to m do
      next[i] := gamma * simp[h,i] + (1.0 - gamma) * center[i];
    sum_of_residuals(next);
    if next[n] <= simp[l,n] then begin
      if trace then write(' e');
      new_vertex;
      end;
    end
  else begin
    if next[n] <= simp[o,n] then begin
      if trace then write(' R');
      new_vertex;
      end
    else if next[n] <= simp[h,n] then begin
        for i := 1 to m do
          next[i] := - beta * simp[h,i] + (1.0 + beta) * center[i];
        sum_of_residuals(next);
        if trace then write('-c');
        new_vertex;
        end
      else begin
        for i := 1 to m do
          next[i] := beta * simp[h,i] + (1.0 - beta) * center[i];
        sum_of_residuals(next);
        if trace then write(' c');
        new_vertex;
        end
      end;
    order;
    error := (simp[h,n] - simp[l,n]) / simp[h,n];
  until ((abs(error) < maxerr) or (niter = maxiter) or (exitcode <> 0));

for i := 1 to n do begin
  mean[i] := 0.0;
  for j := 1 to n do mean[i] := mean[i] + simp[j,i];
  mean[i] := mean[i]/n
  end;
exitproc := exitsave;
if track then close(dout);
end;
{



.pa}
procedure jknife;
const track = true;
var i,j,k,l : integer;
    jdata   : dataset;
    success : integer;
    sdev    : vector;

begin
if track then
  begin
  assign(dout,'jack.res');
{$I-}
  erase(dout);
  rewrite(dout);
  close(dout);
  i := ioresult;
{$I+}
  append(dout);
  end;

simplex(f,false,nvpp,m,np,maxiter,maxerr,data,mean,niter);

if trace then
    begin
    writeln(' best fit:');
    for i := 1 to m+1 do
      writeln(mean[i]);
    writeln('it took ',niter:4,' iterations');
    end;

for i := 1 to m+1 do sdev[i] := 0;
success := np;
for i := 1 to np do
  begin
  k := 1;
  for j := 1 to np do
    begin
    if j <> i then
      begin
      jdata[k] := data[j];
      k := succ(k);
      end
    end;
  error := mean;
  simplex(f,false,nvpp,m,np-1,maxiter,maxerr,jdata,error,niter);
  if niter = maxiter then success := success - 1
  else for l := 1 to m+1 do sdev[l] := sdev[l] + sqr(mean[l]-error[l]);
  if track then
    begin
    for l := 1 to m+1 do
      write(dout,error[l]:13,' ');
    writeln(dout);
    end;
  if trace then
    begin
    writeln('jackknifed set #',i:3,' : parameters are');
    for l := 1 to m+1 do writeln(error[l]);
    writeln('it took ',niter:4,' iterations');
    end;
  end;
  for i := 1 to m+1 do error[i] := sqrt(sdev[i]*(success-1)/np);
  niter := success;
  if track then close(dout);
end;
{



.pa}
procedure bstrap;
const track = true;
var i,j,k,l : integer;
    ntests  : longint;
    mdata   : dataset;
    sdev    : vector;
    success : longint;

begin
randomize;
if track then
  begin
  assign(dout,'boot.res');
{$I-}
  erase(dout);
  rewrite(dout);
  close(dout);
  i := ioresult;
{$I+}
  append(dout);for i := 1 to m+1 do sdev[i] := 0;
  end;
ntests := niter; success := ntests;

simplex(f,false,nvpp,m,np,maxiter,maxerr,data,mean,niter);

if trace then
    begin
    writeln(' best fit:');
    for i := 1 to m+1 do
      writeln(mean[i]);
    writeln('it took ',niter:4,' iterations');
    end;

for i := 1 to ntests do
  begin
  for j := 1 to np do
    begin
    k := trunc(random * np + 1);
    mdata[j] := data[k];
    end;
  error := mean;
  simplex(f,false,nvpp,m,np,maxiter,maxerr,mdata,error,niter);
  if niter = maxiter then success := success - 1
  else for l := 1 to m+1 do sdev[l] := sdev[l] + sqr(mean[l]-error[l]);
  if track then
    begin
    for l := 1 to m+1 do
      write(dout,error[l]:13,' ');
    writeln(dout);
    end;
  if trace then
    begin
    writeln('bootstrapped set #',i:4,' of ',ntests:4,' : parameters are');
    for l := 1 to m+1 do writeln(error[l]);
    writeln('it took ',niter:4,' iterations');
    end;
  end;
  for i := 1 to m+1 do error[i] := sqrt(sdev[i]/(success-m));
  niter := success;
  if track then close(dout);
  end ;
{



.pa}

procedure mcarlo;
const track = true;
var i,l     : integer;
    j,ntests: longint;
    mdata,
    tdata   : dataset;
    sdev    : vector;
    success : longint;
    sigma   : extended;

function grandom : extended;
{to compute random numbers, gaussian distribution, center 0,  sigma 1}
{for additional safety, numbers more than maxr sigmas out are skipped}

const maxr = 3.0;
var t : extended;
  begin
  t := sqrt(-0.868589 * ln(random))*(cos(6.28318531*random));
  if abs(t) < maxr then
    grandom := t
    else grandom := grandom;
  end;

begin
randomize;
if track then
  begin
  assign(dout,'monte.res');
{$I-}
  erase(dout);
  rewrite(dout);
  close(dout);
  i := ioresult;
{$I+}
  append(dout);
  end;
for i := 1 to m+1 do sdev[i] := 0;
ntests := niter; success := ntests;

simplex(f,false,nvpp,m,np,maxiter,maxerr,data,mean,niter);

if trace then
  begin
  writeln(' best fit:');
  for i := 1 to m+1 do
    writeln(mean[i]);
  writeln('it took ',niter:4,' iterations');
  end;

sigma := 0;
for i := 1 to np do
  begin
  for j := 1 to nvpp-1 do mdata[i,j] := data[i,j];
  tdata[i,nvpp] := f(mean,data[i]);
  sigma := sigma + sqr(tdata[i,nvpp]-data[i,nvpp])
  end;
sigma := sqrt(sigma / (np - m - 1));

for j := 1 to ntests do
  begin
  for i := 1 to np do
    mdata[i,nvpp] := tdata[i,nvpp] + grandom * sigma;
  error := mean;
  simplex(f,false,nvpp,m,np,maxiter,maxerr,mdata,error,niter);
  if niter = maxiter then success := success - 1
  else for l := 1 to m+1 do sdev[l] := sdev[l] + sqr(mean[l]-error[l]);
  if track then
    begin
    for l := 1 to m+1 do
      write(dout,error[l]:13,' ');
    writeln(dout);
    end;
  if trace then
    begin
    writeln('montecarlo set #',j:4,' of ',ntests:4,' : parameters are');
    for l := 1 to m+1 do writeln(error[l]);
    writeln('it took ',niter:4,' iterations');
    end;
  end;

  for i := 1 to m+1 do error[i] := sqrt(sdev[i]/(success-m));
  niter := success;
  if track then close(dout);
  end ;

end .
